/***************************************************************************
 *   Copyright (C) 2009 by Stanson <stanson@pangolin.ru>                   *
 *   GPL v3 code, read COPYING file                                        *
 **************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <signal.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <getopt.h>

#include <usb.h>

// 
#define READ_VERSION    0x00
#define READ_FLASH      0x01
#define WRITE_FLASH     0x02
#define ERASE_FLASH     0x03
#define READ_EEDATA     0x04
#define WRITE_EEDATA    0x05
#define READ_CONFIG     0x06
#define WRITE_CONFIG    0x07
#define RESET           0xFF

struct _processor 
{
	char		*name;
	unsigned long	flash;
	unsigned long	eeprom;
	unsigned long	config;
	unsigned long	id;
	unsigned short	dev_id;
};

struct _packet
{
	unsigned char	cmd;
	unsigned char	len;
	unsigned char	adr[3];
	unsigned char	buf[32];
};

struct _memory
{
	int		len;
	unsigned long	addr;
	unsigned char	buf[32];
	struct _memory	*next;
};

void init_packet(struct _packet *pkt, unsigned char cmd, unsigned char len, unsigned long addr)
{
	memset(pkt, 0, sizeof(struct _packet));
	pkt->cmd = cmd;
	pkt->len = len;
	*(unsigned long *)&pkt->adr = addr;
}

int read_byte(FILE *fd, int *sum)
{int ret, byte, i;
	for(byte = 0, i = 0; i < 2; i++)
	{
		byte <<= 4;
		ret = fgetc(fd);
		if(('0' <= ret) && (ret <= '9'))	byte |= ret - '0';
		else if(('A' <= ret) && (ret <= 'F'))	byte |= ret - 'A' + 10;
		else if(('a' <= ret) && (ret <= 'f'))	byte |= ret - 'a' + 10;
		else					return EOF;
	}
	*sum += byte;
	return byte;
}

int read_hex_line(FILE *fd, unsigned long *addr, unsigned char *buf, int *len)
{int ret, sum = 0, i, type;
 static int line = 0;

	if(fgetc(fd) != ':') goto syntax_error;
	if((*len = read_byte(fd, &sum)) == EOF) goto syntax_error;
	if((ret  = read_byte(fd, &sum)) == EOF) goto syntax_error;
	*addr = ret << 8;
	if((ret  = read_byte(fd, &sum)) == EOF) goto syntax_error;
	*addr |= ret;
	if((type = read_byte(fd, &sum)) == EOF) goto syntax_error;
	for(i = 0; i < *len; i++)
	{
		if((ret = read_byte(fd, &sum)) == EOF) goto syntax_error;
		buf[i] = ret;
	}
	if(read_byte(fd, &sum) == EOF) goto syntax_error;
	if(sum & 0xFF) goto checksum_error;
	do { ret = fgetc(fd); } while(ret != EOF && ret != '\n');
	return type;

syntax_error:
	fprintf(stderr, "Syntax error al line: %d\n", line);
	return EOF;

checksum_error:
	fprintf(stderr, "Checksum error al line: %d\n", line);
	return EOF;
}

int add_chunk(struct _memory *mem, unsigned long addr, unsigned char *buf, int len)
{int start, size;


	while(mem->next && len)
	{
		if((mem->addr <= addr) && (addr < mem->addr + mem->len))
		{
			start = addr - mem->addr;
			size  = mem->len - start;
			if(size > len) size = len;
			memcpy(&mem->buf[start], buf, size);
			len  -= size;
			addr += size;
		}
		mem = mem->next;
	}

	if(len)
	{	/*     */
		if(addr < 0x200000)
		{	/*      32  
			 *     32-  */
			mem->addr = addr & 0xFFFFFFE0;
			mem->len  = 32;
		}
		else
		{	/*       
			 *      */
			mem->addr = addr;
			mem->len = (len < 32) ? len : 32;
		}
		memset(mem->buf, 0xFF, 32);
		/*    */
		mem->next = malloc(sizeof(struct _memory));
		if(!mem->next) goto error;
		memset(mem->next, 0, sizeof(struct _memory));
		/*        */
		add_chunk(mem, addr, buf, len);
	}
	return 1;

error:
	fprintf(stderr, "Allocation error!\n");
	return 0;
}

void free_chunks(struct _memory *mem)
{
	/*     ,   
	 * -   */
	if(mem->next) free_chunks(mem->next);
	free(mem);
}

int read_hex(FILE *fd, struct _memory *mem)
{unsigned char buf[256];
 int len, ret;
 unsigned long offset = 0, addr = 0;

	while((ret = read_hex_line(fd, &addr, buf, &len)) != EOF)
	{
		switch(ret)
		{
			case 0: /* DATA */
				if(!add_chunk(mem, offset + addr, buf, len)) return 0;
				break;
			case 1: /* EOF */
				return 1;
			case 4: /* LBA */
				offset = (buf[0] << 24) | (buf[1] << 16);
				break;
		}
	}
	return 0;
}

void write_byte(FILE *fd, unsigned char byte, int *sum)
{
	*sum += byte;
	fprintf(fd, "%02X", byte);
}

void write_hex_line(FILE *fd, unsigned char type, unsigned long addr, unsigned char *data, int len)
{int sum = 0;
	fprintf(fd, ":");
	write_byte(fd, len, &sum);
	write_byte(fd, (addr & 0xFF00) >> 8, &sum);
	write_byte(fd, addr & 0xFF, &sum);
	write_byte(fd, type, &sum);
	while(len--) write_byte(fd, *data++, &sum);
	fprintf(fd, "%02X", (-sum) & 0xFF);
	fprintf(fd, "\n");
}

void write_hex(FILE *fd, unsigned long addr, unsigned char *data, int len)
{int count, i;
 unsigned char buf[2];
 static unsigned long old_addr = 0xFFFFFFFF;
 
	do
	{
		if(old_addr != (addr & 0xFFFF0000))
		{
			old_addr = addr & 0xFFFF0000;
			buf[0] = (addr & 0xFF000000) >> 24;
			buf[1] = (addr & 0x00FF0000) >> 16;
			write_hex_line(fd, 4, addr, buf, 2);
		}
		count = (len > 16) ? 16 : len;
		i = 0;
		/*  EEPROM, ID  CONFIG    */
		if(addr < 0x200000)
		{
			while(i < count) if(data[i++] != 0xFF) break;
		}
		if(i < count) write_hex_line(fd, 0, addr, data, count);
		addr += count;
		data += count;
		len -= count;
	} while(len > 0);
}

int main(int argc, char *argv[])
{int			ret;
 unsigned long		start = 0;
 char			*prog, *filename = 0;
 unsigned char		cmd = 0;
 unsigned short		dev_id;
 static struct option long_options[] =
 {
   {"version", no_argument,       0, 'v'},
   {"device",  required_argument, 0, 'd'},
   {"read",    no_argument,       0, 'r'},
   {"erase",   no_argument,       0, 'e'},
   {"write",   required_argument, 0, 'w'},
   {"list",    no_argument,       0, 'l'},
   {"help",    no_argument,       0, 'h'},
   {0, 0, 0, 0}
 };
 static struct _processor proc[] =
 {
   {"PIC18F2455", 24576, 256, 0x300000, 0x200000, 0x1260},
   {"PIC18F2550", 32768, 256, 0x300000, 0x200000, 0x1240},
   {"PIC18F4455", 24576, 256, 0x300000, 0x200000, 0x1220},
   {"PIC18F4550", 32768, 256, 0x300000, 0x200000, 0x1200},
   {0, 0, 0, 0}
 };
 static struct _processor *p = proc;
 struct _packet pkt;
 struct _memory *mem = 0, *m = 0;
 FILE *file;

    int			iface = 0;
    char		*product, *vendor, *id = "04d8:000b";
    unsigned short	vendor_id  = 0x04d8;
    unsigned short	product_id = 0x000b;
    struct usb_bus 	*usbbus;
    struct usb_device	*usbdev;
    usb_dev_handle	*dev;


	prog = strrchr(argv[0],'/');
	if(!prog) prog = argv[0];
	else      prog++;

	while ((ret = getopt_long (argc, argv, "d:vrew:lh", long_options, NULL)) != -1)
	{
		switch(ret)
		{
    			case 'd':
			    id = strdup(optarg);
			    product = optarg;
    			    vendor = strsep(&product, ":");
			    if(vendor && *vendor) {
				vendor_id = strtol(vendor, NULL, 16);
			    }
			    if(product && *product) {
				product_id = strtol(product, NULL, 16);
			    }
			    break;
			case 'w': cmd = WRITE_FLASH; filename = strdup(optarg); break;
			case 'r': cmd = READ_FLASH; break;
			case 'e': cmd = ERASE_FLASH; break;
			case 'v': cmd = READ_VERSION; break;
			case 'l':
				printf("Supported processors:\n");
				for(p = proc; p->name; p++)
				{
					printf("%s ", p->name);
					printf("flash: %lu bytes, ", p->flash);
					printf("eeprom: %lu bytes, ", p->eeprom);
					printf("config at: %#06lx\n", p->config);
				}
				return 1;
			case 'h':
			default:
				printf("Usage: %s [--list/-l] [--read/-r] [--erase/-e] [--write filename/-w filename] [--help/-h] [--version/-v]\n", prog);
				printf("Options:\n");
				printf("\t-d, --device vendor_id:product_id  Use device with vendor:product id\n");
				printf("\t-l, --list                         List known processors\n");
				printf("\t-r, --read                         Read processor memory\n");
				printf("\t-e, --erase                        Erase processor memory\n");
				printf("\t-w, --write filename               Write Intel HEX file into memory\n");
				printf("\t-v, --version                      Print bootloader version\n");
				printf("\t-h, --help                         Show this help\n");
				return 1;
		}
	}

    usb_init();
    usb_find_busses();
    usb_find_devices();

    usbbus = usb_get_busses();
    for (; usbbus; usbbus = usbbus->next)
    {
	for (usbdev = usbbus->devices; usbdev; usbdev=usbdev->next)
	{
	    if (usbdev->descriptor.idVendor == vendor_id &&
	        usbdev->descriptor.idProduct == product_id) 
	    {
		dev = usb_open(usbdev);
		if(!dev)
		{
		    fprintf(stderr, "Failed to open USB device %s\n", id);
		    ret = -1;
		    goto out;
		}
		ret = usb_claim_interface(dev, iface);
		if(ret < 0)
		{
		    ret = usb_detach_kernel_driver_np(dev, iface);
		    if(ret < 0)
		    {
			fprintf(stderr, "Failed to detach kernel driver from interface %d on device %s\n", iface, id);
			goto out_close;
		    }
		    ret = usb_claim_interface(dev, iface);
		    if(ret < 0)
		    {
			fprintf(stderr, "Failed to claim interface %d on device %s\n", iface, id);
			goto out_close;
		    }
		}
	        goto device_found;
	    }
        }
    }

    fprintf(stderr, "Device %s not found!\n", id);
    goto out;

device_found:

	/*    */
	init_packet(&pkt, READ_FLASH, 2, 0x3FFFFE);

	usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 5, 1000);
	usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, pkt.len + 5, 1000);

	dev_id = (pkt.buf[1] << 8) | pkt.buf[0];
	for(p = proc; p->name; p++)
	{
		if((dev_id & 0xFFE0) == p->dev_id) break;
	}
	if(!p->name)
	{
		fprintf(stderr, "Unsupported processor! Device ID: %#04x\n", dev_id);
		return 1;
	}
	fprintf(stderr, "%s Rev. %d detected\n", p->name, dev_id & 0x001F);

	/*    */
	init_packet(&pkt, READ_VERSION, 0, 0);

	usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 1, 1000);
	usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, 4, 1000);

	fprintf(stderr,"Bootloader version: %x.%x\n",pkt.adr[1],pkt.adr[0]);

	switch(cmd)
	{
		case READ_FLASH:
			/*    */
			start = 0;
			while(start < p->flash)
			{
				init_packet(&pkt, READ_FLASH, 32, start);

				usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 5, 1000);
				usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, pkt.len + 5, 1000);

				write_hex(stdout, start, pkt.buf, pkt.len);
				start += pkt.len;
			}
			/*  ID */
			init_packet(&pkt, READ_CONFIG, 8, p->id);

			usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 5, 1000);
			usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, pkt.len + 5, 1000);

			write_hex(stdout, p->id, pkt.buf, pkt.len);
			/*   */
			init_packet(&pkt, READ_CONFIG, 14, p->config);

			usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 5, 1000);
			usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, pkt.len + 5, 1000);

			write_hex(stdout, p->config, pkt.buf, pkt.len);
			/*    */
			start = 0;
			while(start < p->eeprom)
			{
				init_packet(&pkt, READ_EEDATA, 32, start);

				usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 5, 1000);
				usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, pkt.len + 5, 1000);

				write_hex(stdout, start + 0xF00000, pkt.buf, pkt.len);
				start += pkt.len;
			}
			write_hex_line(stdout, 1, 0, NULL, 0);
			break;
		case ERASE_FLASH:
			/*    */
			start = 0x800;
			while(start < p->flash)
			{
				init_packet(&pkt, ERASE_FLASH, 1, start);

				usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, 5, 1000);
				usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, 1, 1000);

				fprintf(stderr, "Erasing block %#06lx\n", start);
				start += 64;
			}
			break;
		case WRITE_FLASH:
			/*  HEX-   */
			mem = malloc(sizeof(struct _memory));
			if(!mem)
			{
				fprintf(stderr, "Can't allocate memory!\n");
				break;
			}
			if(!strcmp(filename,"-"))
			{
				file = stdin;
			}
			else
			{
				file = fopen(filename,"r");
				if(!file)
				{
					fprintf(stderr, "Can't open file %s. %s\n", filename, strerror(errno));
					break;
				}
			}

			memset(mem, 0, sizeof(struct _memory));
			if(read_hex(file, mem))
			{

				m = mem;
				while(m->next)
				{
					if((0x00000800 <= m->addr) && (m->addr < 0x200000))
					{
						init_packet(&pkt, WRITE_FLASH, m->len, m->addr);
						memcpy(pkt.buf, m->buf, m->len);
						usb_bulk_write(dev, USB_ENDPOINT_OUT | 1, (char *)&pkt, m->len + 5, 1000);
						usb_bulk_read(dev, USB_ENDPOINT_IN | 1, (char *)&pkt, 1, 1000);
						fprintf(stderr, "Writing block %#06lx %d bytes\n", m->addr, m->len);
					}
					m = m->next;
				}
			}
			free_chunks(mem);
			if(file != stdin) fclose(file);
			free(filename);
			break;
	}

out_close:
    usb_close(dev);
out:
    return ret;
}
